package edu.uiuc.ncsa.myproxy.oa4mp.server;

import edu.uiuc.ncsa.myproxy.MyProxyServiceFacade;
import edu.uiuc.ncsa.myproxy.oa4mp.server.servlet.AuthorizationServlet;
import edu.uiuc.ncsa.myproxy.oa4mp.server.servlet.AuthorizationServletConfig;
import edu.uiuc.ncsa.myproxy.oa4mp.server.util.AbstractCLIApprover;
import edu.uiuc.ncsa.security.core.configuration.Configurations;
import edu.uiuc.ncsa.security.core.configuration.provider.CfgEvent;
import edu.uiuc.ncsa.security.core.configuration.provider.HierarchicalConfigProvider;
import edu.uiuc.ncsa.security.core.exceptions.GeneralException;
import edu.uiuc.ncsa.security.core.util.AbstractEnvironment;
import edu.uiuc.ncsa.security.core.util.MyLoggingFacade;
import edu.uiuc.ncsa.security.core.util.QueueWithSpare;
import edu.uiuc.ncsa.security.delegation.server.issuers.AccessTokenIssuer;
import edu.uiuc.ncsa.security.delegation.server.issuers.AuthorizationGrantIssuer;
import edu.uiuc.ncsa.security.delegation.server.issuers.ProtectedAssetIssuer;
import edu.uiuc.ncsa.security.delegation.server.storage.ClientApproval;
import edu.uiuc.ncsa.security.delegation.server.storage.ClientApprovalStore;
import edu.uiuc.ncsa.security.delegation.server.storage.ClientStore;
import edu.uiuc.ncsa.security.delegation.storage.TransactionStore;
import edu.uiuc.ncsa.security.delegation.token.TokenForge;
import edu.uiuc.ncsa.security.servlet.mail.MailUtil;
import edu.uiuc.ncsa.security.servlet.mail.MailUtilProvider;
import edu.uiuc.ncsa.security.util.pkcs.KeyUtil;
import org.apache.commons.configuration.tree.ConfigurationNode;

import javax.inject.Provider;
import java.net.URI;
import java.security.KeyPair;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import static edu.uiuc.ncsa.myproxy.oa4mp.server.servlet.AuthorizationServlet.RETRY_MESSAGE;

/**
 * The runtime environment for a service. This is a bridge between the configuration and the instances.
 * Typically it is populated with providers (i.e.configured factories) which retrieves the instances
 * as needed.
 * <p>Created by Jeff Gaynor<br>
 * on 1/9/12 at  4:08 PM
 */
public class ServiceEnvironmentImpl extends AbstractEnvironment implements ServiceEnvironment {
    public KeyPairQueue getKeyPairQueue() {
        return kpq;
    }

    KeyPairQueue kpq = new KeyPairQueue();

    public KeyPair getKeyPair() {
        return kpq.pop();
    }

    public static class KeyPairQueue extends QueueWithSpare<KeyPair> {
        @Override
        public KeyPair pop() {
            KeyPair x = super.pop();
            if (x != null) {
                return x;
            }
            try {
                KeyPair keyPair = KeyUtil.generateKeyPair();
                setSpare(keyPair);
                return keyPair;
            } catch (Exception e) {
                throw new GeneralException("Error generating key pair", e);
            }

        }
    }

    AuthorizationServletConfig authorizationServletConfig;

    public AuthorizationServletConfig getAuthorizationServletConfig() {
        return authorizationServletConfig;
    }

    AuthorizationGrantIssuer authorizationGrantIssuer;
    URI serviceAddress;

    @Override
    public URI getServiceAddress() {
        return serviceAddress;
    }

    public void setServiceAddress(URI serviceAddress) {
        this.serviceAddress = serviceAddress;
    }

    @Override
    public AuthorizationGrantIssuer getAuthorizationGrantIssuer() {
        if (authorizationGrantIssuer == null) {
            authorizationGrantIssuer = agip.get();
        }
        return authorizationGrantIssuer;
    }

    AccessTokenIssuer accessTokenIssuer;

    @Override
    public AccessTokenIssuer getAccessTokenIssuer() {
        if (accessTokenIssuer == null) {
            accessTokenIssuer = atip.get();
        }
        return accessTokenIssuer;
    }

    @Override
    public ProtectedAssetIssuer getProtectedAssetIssuer() {
        if (protectedAssetIssuer == null) {
            protectedAssetIssuer = paip.get();
        }
        return protectedAssetIssuer;
    }

    ProtectedAssetIssuer protectedAssetIssuer;

    @Override
    public TokenForge getTokenForge() {
        return tfp.get();
    }

    public static class MessagesProvider extends HierarchicalConfigProvider<Map<String, String>> {

        public MessagesProvider(ConfigurationNode config) {
            super(config);
        }

        @Override
        protected boolean checkEvent(CfgEvent cfgEvent) {
            if (cfgEvent.getConfiguration().getName().equals("messages")) {
                setConfig(cfgEvent.getConfiguration());
                return true;
            }
            return false;
        }

        @Override
        public Object componentFound(CfgEvent configurationEvent) {
            if (checkEvent(configurationEvent)) {
                return get();
            }
            return null;
        }

        @Override
        public Map<String, String> get() {
            HashMap<String, String> messages = new HashMap<String, String>();
            messages.put(RETRY_MESSAGE, Configurations.getNodeValue(getConfig(), RETRY_MESSAGE, "Authentication failed."));
            return messages;
        }
    }


    public ServiceEnvironmentImpl(MyLoggingFacade logger,
                                  Provider<TransactionStore> tsp,
                                  Provider<ClientStore> csp,
                                  int maxAllowedNewClientRequests,
                                  Provider<ClientApprovalStore> casp,
                                  List<MyProxyFacadeProvider> mfp,
                                  MailUtilProvider mup,
                                  MessagesProvider messagesProvider,
                                  Provider<AuthorizationGrantIssuer> agip,
                                  Provider<AccessTokenIssuer> atip,
                                  Provider<ProtectedAssetIssuer> paip,
                                  Provider<TokenForge> tfp,
                                  HashMap<String,String> constants,
                                  AuthorizationServletConfig ac) {
        super(logger, constants);
        this.casp = casp;
        this.csp = csp;
        this.tsp = tsp;
        this.mfps = mfp;
        this.mup = mup;
        this.messagesProvider = messagesProvider;
        this.atip = atip;
        this.agip = agip;
        this.paip = paip;
        this.tfp = tfp;
        this.authorizationServletConfig = ac;
        this.maxAllowedNewClientRequests = maxAllowedNewClientRequests;
    }

    MessagesProvider messagesProvider;
    MailUtilProvider mup;


    protected Provider<TransactionStore> tsp;
    protected Provider<ClientStore> csp;
    protected Provider<ClientApprovalStore> casp;
    protected Provider<AuthorizationGrantIssuer> agip;
    protected Provider<AccessTokenIssuer> atip;
    protected Provider<ProtectedAssetIssuer> paip;
    protected Provider<TokenForge> tfp;



    Map<String, String> messages;

    @Override
    public Map<String, String> getMessages() {
        if (messages == null) {
            if (messagesProvider == null) {
                messages = new HashMap<String, String>();
                messages.put(AuthorizationServlet.RETRY_MESSAGE, "Authentication failed.");
            } else {
                messages = messagesProvider.get();
            }
        }
        return messages;
    }

    @Override
    public ClientApprovalStore getClientApprovalStore() {
        if (clientApprovalStore == null) {
            clientApprovalStore = casp.get();
        }
        return clientApprovalStore;
    }

    @Override
    public ClientStore getClientStore() {
        if (clientStore == null) {
            clientStore = csp.get();
        }
        return clientStore;
    }

    protected MailUtil mailUtil;
    protected ClientApprovalStore<ClientApproval> clientApprovalStore;
    protected ClientStore clientStore;
    protected TransactionStore<DSTransaction> transactionStore;
    protected List<MyProxyServiceFacade> myProxyServices;

    @Override
    public MailUtil getMailUtil() {
        if (mailUtil == null) {
            mailUtil = mup.get();
        }
        return mailUtil;
    }

    List<MyProxyFacadeProvider> mfps;

    @Override
    public List<MyProxyServiceFacade> getMyProxyServices() {
        if (myProxyServices == null) {
            myProxyServices = new LinkedList<MyProxyServiceFacade>();
            // loop through each found component
            for (MyProxyFacadeProvider m : mfps) {
                myProxyServices.add(m.get());
            }
            return myProxyServices;
        }
        return myProxyServices;
    }


    @Override
    public TransactionStore<DSTransaction> getTransactionStore() {
        if (transactionStore == null) {
            transactionStore = (TransactionStore<DSTransaction>) tsp.get();
        }
        return transactionStore;
    }

    public boolean isPollingEnabled() {
        return clientApprovalThread != null;
    }

    public AbstractCLIApprover.ClientApprovalThread getClientApprovalThread() {
        return clientApprovalThread;
    }

    public void setClientApprovalThread(AbstractCLIApprover.ClientApprovalThread clientApprovalThread) {
        this.clientApprovalThread = clientApprovalThread;
    }

    AbstractCLIApprover.ClientApprovalThread clientApprovalThread;

    int maxAllowedNewClientRequests = 100;

    @Override
    public int getMaxAllowedNewClientRequests() {
        return maxAllowedNewClientRequests;
    }
}
