package edu.uiuc.ncsa.myproxy.oa4mp.server.servlet;

import edu.uiuc.ncsa.myproxy.oa4mp.server.util.NewClientEvent;
import edu.uiuc.ncsa.myproxy.oa4mp.server.util.NewClientListener;
import edu.uiuc.ncsa.security.core.Identifier;
import edu.uiuc.ncsa.security.delegation.server.storage.ClientApproval;
import edu.uiuc.ncsa.security.delegation.servlet.TransactionState;
import edu.uiuc.ncsa.security.delegation.storage.Client;
import edu.uiuc.ncsa.security.servlet.JSPUtil;
import edu.uiuc.ncsa.security.servlet.NotificationListener;
import edu.uiuc.ncsa.security.servlet.Presentable;
import edu.uiuc.ncsa.security.servlet.PresentableState;
import edu.uiuc.ncsa.security.util.pkcs.KeyUtil;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.util.HashMap;

/**
 * The servlet that handles registration. This will create the correct state then forward the request to the correct JSP
 * page. Since this implements the {@link Presentable} interface, you may over-ride it to display information anyway you wish.
 * <p>Created by Jeff Gaynor<br>
 * on 9/25/11 at  3:53 PM
 */
public class RegistrationServlet extends MyProxyDelegationServlet implements Presentable {
    public static final String CLIENT_NAME = "clientName";
    public static final String CLIENT_PUBLIC_KEY = "clientPublicKey";
    public static final String CLIENT_HOME_URL = "clientHomeUrl";
    public static final String CLIENT_ERROR_URL = "clientErrorUrl";
    public static final String CLIENT_EMAIL = "clientEmail";
    public static final String CLIENT_PROXY_LIMITED = "clientProxyLimited";
    public static final String CLIENT_ACTION_KEY = "action";
    public static final String CLIENT_ACTION_REQUEST_VALUE = "request";

    protected static final int INITIAL_STATE = 0;
    protected static final int ERROR_STATE = -1;
    protected static final int REQUEST_STATE = 100;

    public int getState(HttpServletRequest request) {
        String action = request.getParameter(CLIENT_ACTION_KEY);
        if (action == null) return INITIAL_STATE;
        if (action.equals(CLIENT_ACTION_REQUEST_VALUE)) return REQUEST_STATE;
        return ERROR_STATE; // something is wrong with the request or state,
    }

    protected void fireNewClientEvent(Client client) {
        for (NotificationListener notificationListener : notificationListeners) {
            if (notificationListener instanceof NewClientListener) {
                ((NewClientListener) notificationListener).fireNewClientEvent(new NewClientEvent(this, client));
            }
        }
    }

    public void prepare(PresentableState state) throws Throwable {
        preprocess(new TransactionState(state.getRequest(), state.getResponse(), null, null));
        switch (state.getState()) {
            case INITIAL_STATE:
                HttpServletRequest request = state.getRequest();
                info("Processing new client registration request.");
                request.setAttribute(CLIENT_NAME, CLIENT_NAME);
                request.setAttribute(CLIENT_PUBLIC_KEY, CLIENT_PUBLIC_KEY);
                request.setAttribute(CLIENT_HOME_URL, CLIENT_HOME_URL);
                request.setAttribute(CLIENT_ERROR_URL, CLIENT_ERROR_URL);
                request.setAttribute(CLIENT_EMAIL, CLIENT_EMAIL);
                request.setAttribute(CLIENT_PROXY_LIMITED, CLIENT_PROXY_LIMITED);
                request.setAttribute(CLIENT_ACTION_KEY, CLIENT_ACTION_KEY);
                request.setAttribute(CLIENT_ACTION_REQUEST_VALUE, CLIENT_ACTION_REQUEST_VALUE);
                request.setAttribute("actionToTake", request.getContextPath() + "/register");
                break;
            case REQUEST_STATE:
                // nothing to do.
                return;
            case ERROR_STATE:
            default:
                warn("Error: unknown action request.");
        }

    }

    /**
     * The page to display to the client for the initial request.
     */
    public static String INIT_PAGE = "/registration-init.jsp";

    /**
     * The name of a JSP page to display in  case of errors. The default is "registration-error.jsp".
     */
    public static String ERROR_PAGE = "/registration-error.jsp";

    /**
     * If the registration works, this is the page to display to the user afterwards.
     */

    public static String OK_PAGE = "/registration-ok.jsp";


    public void present(PresentableState state) throws Throwable {
        postprocess(new TransactionState(state.getRequest(), state.getResponse(), null, null));

        switch (state.getState()) {
            case INITIAL_STATE:
                JSPUtil.fwd(state.getRequest(), state.getResponse(), INIT_PAGE);
                break;
            case REQUEST_STATE:
                if (state instanceof ClientState) {
                    ClientState cState = (ClientState) state;
                    state.getRequest().setAttribute("client", cState.getClient());
                    JSPUtil.fwd(state.getRequest(), state.getResponse(), OK_PAGE);
                } else {
                    throw new IllegalStateException("Error: An instance of ClientState was expected, but got an instance of \"" + state.getClass().getName() + "\"");
                }
                break;
            case ERROR_STATE:
            default:
        }
    }

    public void handleError(PresentableState state, Throwable t) throws IOException, ServletException {
        state.getResponse().setHeader("X-Frame-Options", "DENY");
        JSPUtil.handleException(t, state.getRequest(), state.getResponse(), ERROR_PAGE);
    }

    static class ClientState extends PresentationState {
        ClientState(int state,
                    HttpServletRequest request,
                    HttpServletResponse response,
                    Client client) {
            super(state, request, response);
            this.client = client;
        }

        public Client getClient() {
            return client;
        }

        Client client;
    }

    @Override
    protected void doIt(HttpServletRequest request, HttpServletResponse response) throws Throwable {
        int state = getState(request);
        info("unapproved count = " + getServiceEnvironment().getClientApprovalStore().getUnapprovedCount());
        info("max new client requests = " + getServiceEnvironment().getMaxAllowedNewClientRequests() + ", size= " + getServiceEnvironment().getClientApprovalStore().size());
        // Addresses
        if (state == INITIAL_STATE) {
            if (getServiceEnvironment().getMaxAllowedNewClientRequests() <= getServiceEnvironment().getClientApprovalStore().getUnapprovedCount()) {
             //   throw new TooManyRequestsException("Error: Max number of new client requests reached. Request rejected.");
                       JSPUtil.fwd(request, response, "/tooManyClientRequests.jsp");
            }
        }

        PresentationState pState = new PresentationState(state, request, response);
        try {
            prepare(pState);
            if (state == REQUEST_STATE) {
                Client client = addNewClient(request, response);
                pState = new ClientState(state, request, response, client);
            }
            present(pState);
        } catch (Throwable t) {
            warn("Error registering a new client:" + t.getMessage());
            handleError(pState, t);
        }
    }

    String getRequiredParam(HttpServletRequest req, String key) {
        String x = req.getParameter(key);
        if (x == null || x.length() == 0) {
            throw new IllegalArgumentException("Error: missing parameter for " + key);
        }
        return x;
    }

    boolean getBooleanParam(HttpServletRequest req, String key) {
        String x = req.getParameter(key);
        if (x == null || x.length() == 0) {
            return false;
        }
        return Boolean.parseBoolean(x);

    }

    protected Client addNewClient(HttpServletRequest request, HttpServletResponse response) throws Throwable {
        // Assumption is that the request is in good order and we just have to pull stuff off it.
        Client client = getServiceEnvironment().getClientStore().create();
        info("creating entry for client=" + client.getIdentifierString());
        // Fill in as much info as we can before parsing public key.
        // We always store exactly what was given to us, though later we html escape it to
        // prevent against HTML injection attacks (fixes bug OAUTH-87).
        client.setName(getRequiredParam(request, CLIENT_NAME));
        client.setSecret(getRequiredParam(request, CLIENT_PUBLIC_KEY));
        client.setErrorUri(getRequiredParam(request, CLIENT_ERROR_URL));
        client.setHomeUri(getRequiredParam(request, CLIENT_HOME_URL));
        client.setEmail(getRequiredParam(request, CLIENT_EMAIL));
        client.setProxyLimited(getBooleanParam(request, CLIENT_PROXY_LIMITED));
        try {
            debug("decoding public key from PEM");
            KeyUtil.fromX509PEM(client.getSecret());
        } catch (Throwable t) {
            warn("could not decode public key for client=" + client.getIdentifierString() + ", message:" + t.getMessage());
            request.setAttribute("client", client);
            throw t;
        }

        getServiceEnvironment().getClientStore().save(client);
        info("Adding approval record for client=" + client.getIdentifierString());
        ClientApproval clientApproval = new ClientApproval(client.getIdentifier());
        clientApproval.setApproved(false);

        getServiceEnvironment().getClientApprovalStore().save(clientApproval);

        fireNewClientEvent(client);
        info("done with client registration, client=" + client.getIdentifierString());
        return client;
    }

    /**
     * This will approve a client. Supply the approver and client
     *
     * @param clientIdentifier
     * @param approver
     */
    public static void approveClient(Identifier clientIdentifier, String approver) throws IOException {
        ClientApproval clientApproval = getServiceEnvironment().getClientApprovalStore().get(clientIdentifier);
        if (approver == null) {
            approver = ""; // so you don't get something ugly in the backend.
        }
        clientApproval.setApprover(approver);
        clientApproval.setApproved(true);
        getServiceEnvironment().getClientApprovalStore().save(clientApproval);
    }

    protected boolean sendNotification(Client client) throws IOException {
        if (!getServiceEnvironment().getMailUtil().isEnabled()) {
            return true; // everything is always ok if no messages sent
        }
        HashMap<String, String> replacements = new HashMap<String, String>();
        replacements.put("name", client.getName());
        replacements.put("email", client.getEmail());
        replacements.put("homeUri", client.getHomeUri());
        replacements.put("failureUri", client.getErrorUri());
        replacements.put("creationTime", client.getCreationTS().toString());
        replacements.put("identifier", client.getIdentifierString());
        replacements.put("limitedProxy", Boolean.toString(client.isProxyLimited()));
        boolean rc = getServiceEnvironment().getMailUtil().sendMessage(replacements);
        if (rc) {
            info("sending email notification for client " + client.getIdentifierString());
        } else {
            info("failure sending email notification for client " + client.getIdentifierString());
        }
        return rc;
    }


}
